#include <fenv.h>
#include <iostream>
//#include <fstream>
#include <cstdlib>
#include <ctime>
#include "evaluator.h"
#include "rbfmatrix.h"
#include <sys/time.h>


extern __inline__ unsigned long long int rdtsc()
{
   unsigned long long int x;

   __asm__ __volatile__  (               // serialize
      "xorl %%eax,%%eax\n\tcpuid\n"
      ::: "%eax", "%ebx", "%ecx", "%edx");

   __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));

   __asm__ __volatile__  (               // serialize
      "xorl %%eax,%%eax\n\tcpuid\n"
      ::: "%eax", "%ebx", "%ecx", "%edx");


   return x;
}



using namespace fmi;
using namespace std;

int main( int argc, char *argv[] )
{
//   feenableexcept(FE_INVALID);
//   feenableexcept(FE_DIVBYZERO);
//   feenableexcept(FE_OVERFLOW);

   int exitstatus = 0;

   int numberOfPoints = 200;


   if(argc >= 2)
      numberOfPoints = atoi(argv[1]);

   //typedef float ScalarType;

   typedef std::vector< ScalarType > VectorType;
   

   // Reserve memory for the points and coefficients
   VectorType coefficients(numberOfPoints);
   VectorType pointset(3*numberOfPoints);

//   srand48((long int)time(0));

   double fac = 100.0;
   for(int n=0;n<numberOfPoints;)
   {
      pointset[3*n] = drand48()*fac;
      pointset[3*n+1] = drand48()*fac;
      pointset[3*n+2] = drand48()*fac;


      coefficients[n] = drand48()*fac;
      //coefficients[n] = 1.0;
/*
      if(pointset[3*n]   <= 0.0 || pointset[3*n]   >= 1.0 ||
	 pointset[3*n+1] <= 0.0 || pointset[3*n+1] >= 1.0 ||
	 pointset[3*n+2] <= 0.0 || pointset[3*n+2] >= 1.0)
      {
	 continue;
      }
*/      
      n++;
     
   }


   double elapsedTimeFMM,elapsedTimeDirect;
   //timeval t1,t2;
   unsigned long long t1,t2,overhead=2000000000;
   for(int i=0;i<100;i++)
   {
      t1 = rdtsc();
      t2 = rdtsc() - t1;
      if(t2 < overhead)
	 overhead =  t2;
   }

//   std::cout << overhead << std::endl;

//   std::cout.setf(std::ios_base::scientific,std::ios_base::floatfield);
//   std::cout.precision(16);

   // First do FMM method
   //int pp=6;
   for(int ii=1;ii<=1;ii++)
   {
   

      //int dep=3;
      int pp=10;

      std::cout << "\nStarting FMM Multiplication\n" << std::endl;
      cout << "P: " << pp << endl;


	
      unsigned int start_time, end_time;
      double elapsed_time;
      VectorType result(numberOfPoints);
	
      evaluator eval;


      eval.SetPointSet( &pointset[0], numberOfPoints );
      eval.SetCoefficients( &coefficients[0], numberOfPoints );
      //ScalarType extent[6] = {0, 1, 0, 1, 0, 1};
      //eval.SetBoundingBox(extent);
      eval.SetExtraPolynomial( -1,0 );
      eval.SetP(pp);
      eval.SetK(1);
      //eval.SetTau( std::pow(1.0/numberOfPoints,1.0/3.0) );
      eval.SetTau( 0.0 );
      eval.SetMinDegree(6);
//      eval.SetWellSeparatedFactor(2.0);
//      eval.SetDecreaseFactor(1.85);
//      eval.SetDecreaseFactor2(-0.85);
//      eval.SetWellSeparatedFactor(2.0);
//      eval.SetDecreaseFactor(0);
//      eval.SetDecreaseFactor2(0);

      eval.SetDepthToOptimal();
      //eval.SetDepth(2);


      start_time = clock();
      t1 = rdtsc();

      eval.Update();

      eval.EvaluateAtAllSourcePoints(&result[0]);

      t2 = rdtsc();
      end_time = clock();

      elapsedTimeFMM = elapsed_time = 
	 (end_time - start_time )/(double)CLOCKS_PER_SEC;
      std::cout << " Run time = " << (t2 - t1 - overhead) << " cycles = ";
      cout << elapsed_time       << " sec  = ";
      cout << (elapsed_time/60.0) << " min = ";
      cout << (elapsed_time/3600.0) << " hr" << endl;

 	 
      double r,r2,average_error = 0.0;
      double max_error = 0.0;
      double min_error = 1e30;

      average_error = 0.0;
      max_error = 0.0;

      for(int i=0;i<numberOfPoints;i++)
      {

	 //p = points[i];

	 //r = eval.Evaluate(p);
	 //r2 = eval.EvaluateDirect(p);

	 // cout << interp.Evaluate(p) << endl;

	 r = result[i];
	 r2 = eval.EvaluateDirect(&pointset[3*i]);
	 //r2 = eval.EvaluateDirect(&pointset[3*eval.m_PointsValues[i].index]);
	 //r3 = points.GetValuesPointer()[i];
	 //cout << r << " " << r3 << endl;
	
	
	 average_error += fabs(r-r2);
	 //average_error2 += fabs(r-r3);

	 if( fabs(r-r2) > max_error )
	    max_error = fabs(r-r2);
	 if( fabs(r-r2) < min_error )
	    min_error = fabs(r-r2);
	 //if( fabs(r-r3) > max_error2 )
	 // max_error2 = fabs(r-r3);
      
      }

      average_error /= numberOfPoints;
      //average_error2 /= numberOfPoints;
	
      cout << endl << "average_error: " << average_error << endl;
      cout << "max_error: " << max_error << endl;
      cout << "min_error: " << min_error << endl;
      //cout << endl << "average_error2: " << average_error2 << endl;
      //cout << "max_error2: " << max_error2 << endl;

      if(average_error > 1e-5)
	 exitstatus = 1;


   } // end for
	
	
   // Now do regular matrix multiplication

   {

      std::cout << "\nStarting Regular Matrix Multiplication" << std::endl;
      unsigned int start_time, end_time;
      double elapsed_time;


      //double tau = std::pow(1.0/numberOfPoints,1.0/3.0);
      double tau = 0.0;
   
      rbfmatrix m_A;
      m_A.SetupMatrix(&pointset[0],numberOfPoints,-1,1,tau);
	
      VectorType m_x(m_A.size1());
      VectorType m_b(m_A.size1());

      for(int i=0;i<numberOfPoints;i++)
	 m_x[i] = coefficients[i];

      start_time = clock();
	
      m_A.mult(&m_x[0],&m_b[0]);
      //m_b = m_A.mult(m_x);
   
      end_time = clock();
      //std::cout << "Ending Regular Matrix Multiplication" << std::endl;
	
      elapsedTimeDirect =elapsed_time =
	 (end_time - start_time )/(double)CLOCKS_PER_SEC;
	
      std::cout << "\n Run time      =  " << elapsed_time << "  sec  =  ";
      cout << (elapsed_time/60.0) << "  min  =  ";
      cout << (elapsed_time/3600.0) << "  hr" << endl;
   }

   if(elapsedTimeFMM)
   {
      std::cout << "\nTime Ratio: " << (elapsedTimeDirect/elapsedTimeFMM) 
		<< std::endl;
   }

   return exitstatus;

}

/*
int main_()
{

   char str[20];

   int pointslist[] = { 2000, 4000, 8000, 16000, 32000, 64000 };
   for(int i=0; i<6; i++)
   {
      sprintf(str,"%d", pointslist[i]);
      main(2,str);
   }

}
*/
